/* Copyright 2011  Woboq UG
   http://woboq.com/blog/introduction-to-lockfree-programming.html
*/

#include <QtCore/QtCore>

enum { NumNode = 2000000,
       NumThread = 4 };


struct Node {
    Node *next;
};

#if 0
class Stack {
    Node *head;
    QMutex mutex;
public:
    Stack() : head(0) {}
    void push(Node *n) {
        QMutexLocker locker(&mutex);
        n->next = head;
        head = n;
    }
    Node *pop() {
        QMutexLocker locker(&mutex);
        Node *n = head;
        if (n)
            head = n->next;
        return n;
    }
};
#else
class Stack {
    QAtomicPointer<Node> head;
public:
    Stack() : head(0) {}
    void push(Node *n) {
        do {
            n->next = head;
        } while(!head.testAndSetOrdered(n->next, n));
    }
    Node *pop() {
        Node *n;
        do {
            n = head;
            if (!n)
                return 0;
        } while(!head.testAndSetOrdered(n, n->next));
        return n;
    }
};
#endif


struct Pusher : QThread {
    Stack *stack;
    void run() {
        for (int f = 0; f < NumNode/NumThread; f++) {
            stack->push(new Node);
        }
    }
};

struct Popper : QThread {
    Stack *stack;
    int count;
    void run() {
        count = 0;
        Node *n = stack->pop();
        while (n) {
            delete n;
            count ++;
            n = stack->pop();
        }
    }
};

int main() {
   
    Stack stack;
    
    {
        QElapsedTimer timer;
        Pusher pushers[NumThread];
        timer.start();
        for (int t = 0; t < NumThread; t++) {
            pushers[t].stack = &stack;
            pushers[t].start();
        }
        for (int t = 0; t < NumThread; t++) {
            pushers[t].wait();
        }
        qDebug() << "Pushed done" << timer.elapsed();
    }
    
    {
        QElapsedTimer timer;
        Popper poppers[NumThread];
        timer.start();
        for (int t = 0; t < NumThread; t++) {
            poppers[t].stack = &stack;
            poppers[t].start();
        }
        int total = 0;
        int diff = 0;
        for (int t = 0; t < NumThread; t++) {
            poppers[t].wait();
            total += poppers[t].count;
            diff += qAbs((NumNode/NumThread) - poppers[t].count);
        }
        qDebug() << "Destroyed" << total  << "node in" << timer.elapsed();
        qDebug() << diff/NumThread;
    }
    return 0;
}
